package httpserver

import (
	"context"
	"crypto/tls"
	"fmt"
	"io"
	"log"
	"net/http"
	"sync"
	"time"

	"github.com/gorilla/handlers"
	"github.com/gorilla/mux"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	"github.com/it-novum/nsta/nstautil"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// WebserverType defines how the webserver will be started
type WebserverType int

const (
	// WebserverTypeMaster will start the webserver with nsta master functions
	WebserverTypeMaster WebserverType = 0
	// WebserverTypeSatellite will start the webserver with nsta satellite functions
	WebserverTypeSatellite WebserverType = 1
)

// SatelliteIO is the interface to the satellite processor
type SatelliteIO interface {
	SatelliteID() int
	SatelliteName() string
	APIKey() string
	Incoming(*gearman.Message)
	Outgoing() []*gearman.Message
	LastExport() time.Time
	Export(msg *httpproto.ExportData) error
	WebStatus(webStatus *satstatus.Status)
}

// RegisterSatelliteIO can be used to access the webserver functions outside of the package
type RegisterSatelliteIO interface {
	RegisterSatelliteIO(SatelliteIO)
	UnregisterSatelliteIO(int)
}

// NSTAHTTPServer provides a webserver for satellite communication
type NSTAHTTPServer struct {
	serverHTTP  *http.Server
	serverHTTPS *http.Server

	certFile string
	keyFile  string

	webserverType WebserverType
	satelliteIO   map[int]SatelliteIO

	nagiosPath string

	stopped bool
	mtx     sync.Mutex
	wg      sync.WaitGroup
}

func (s *NSTAHTTPServer) satcom(w http.ResponseWriter, req *http.Request) {
	remoteAddr := req.RemoteAddr
	msg := &httpproto.HTTPGearmanMessage{
		SatelliteID: -1,
		APIKey:      "",
		LastExport:  time.Time{},
		Messages:    []*gearman.Message{},
	}
	if err := json.NewDecoder(req.Body).Decode(msg); err != nil {
		log.Println(&Error{Name: "Local", Err: err})
		http.Error(w, "internal server error", 500)
		return
	}

	if msg.SatelliteID == -1 || msg.APIKey == "" {
		log.Println(&Error{Name: remoteAddr, Message: "missing required message parameter"})
		http.Error(w, "internal server error", 500)
		return
	}

	sio, ok := s.satelliteIO[msg.SatelliteID]
	if !ok {
		log.Println(&Error{Name: remoteAddr, Message: fmt.Sprint("unknown satellite id ", msg.SatelliteID)})
		http.Error(w, "internal server error", 500)
		return
	}

	if sio.APIKey() != msg.APIKey {
		log.Println(&Error{Name: remoteAddr, Message: "satellite auth failed"})
		http.Error(w, "internal server error", 500)
		return
	}

	status := &satstatus.Status{
		SatelliteID: msg.SatelliteID,
		Status:      oitcdb.SatelliteStatusOK,
		LastSeen:    time.Now(),
	}
	defer sio.WebStatus(status)

	for _, gm := range msg.Messages {
		// Call Incoming func of the Satellite instance
		sio.Incoming(gm)
	}

	lastExport := sio.LastExport() // Call LastExport func of the Satellite instance
	msgs := sio.Outgoing()         // Call Outgoing func of the Satellite instance
	log.Println(&Error{Name: sio.SatelliteName(), Message: fmt.Sprint("received ", len(msg.Messages), " messages and sending ", len(msgs), " messages")})
	err := json.NewEncoder(w).Encode(&httpproto.HTTPGearmanMessage{SatelliteID: msg.SatelliteID, APIKey: msg.APIKey, Messages: msgs, LastExport: lastExport})
	if err != nil {
		status.Err = &Error{Name: sio.SatelliteName(), Message: "could not marshal response", Err: err}
		log.Println(status.Err)
		http.Error(w, "internal server error", 500)
		return
	}
}

// Returns config for satellite as base64 tar.gz string
func (s *NSTAHTTPServer) exportMaster(w http.ResponseWriter, req *http.Request) {
	remoteAddr := req.RemoteAddr
	msg := &httpproto.ExportData{
		SatelliteID: -1,
		APIKey:      "",
		LastExport:  time.Time{},
		TARBase64:   "",
	}
	if err := json.NewDecoder(req.Body).Decode(msg); err != nil {
		log.Println(&Error{Name: "Local", Err: err})
		http.Error(w, "internal server error", 500)
		return
	}

	if msg.SatelliteID == -1 || msg.APIKey == "" {
		log.Println(&Error{Name: remoteAddr, Message: "missing required message parameter"})
		http.Error(w, "internal server error", 500)
		return
	}

	sio, ok := s.satelliteIO[msg.SatelliteID]
	if !ok {
		log.Println(&Error{Name: remoteAddr, Message: fmt.Sprint("unknown satellite id ", msg.SatelliteID)})
		http.Error(w, "internal server error", 500)
		return
	}

	if sio.APIKey() != msg.APIKey {
		log.Println(&Error{Name: remoteAddr, Message: "satellite auth failed"})
		http.Error(w, "internal server error", 500)
		return
	}

	lastExport := sio.LastExport()

	response := &httpproto.ExportData{
		SatelliteID: sio.SatelliteID(),
		APIKey:      sio.APIKey(),
		LastExport:  lastExport,
		TARBase64:   "",
	}

	response.TARBase64 = nstautil.CreateExportTar(s.nagiosPath, msg.SatelliteID, true)

	if err := json.NewEncoder(w).Encode(response); err != nil {
		log.Println(&Error{Name: "Local", Message: "could not marshal response", Err: err})
		http.Error(w, "internal server error", 500)
		return
	}

	sio.WebStatus(&satstatus.Status{
		SatelliteID: msg.SatelliteID,
		LastExport:  lastExport,
	})
}

// Runs on the satellite and process the export sent by pull satellite instance on the main server via httpclient
func (s *NSTAHTTPServer) exportSatellite(w http.ResponseWriter, req *http.Request) {
	remoteAddr := req.RemoteAddr
	msg := &httpproto.ExportData{
		SatelliteID: -1,
		APIKey:      "",
		LastExport:  time.Time{},
		TARBase64:   "",
	}
	if err := json.NewDecoder(req.Body).Decode(msg); err != nil {
		log.Println(&Error{Name: "Local", Err: err})
		http.Error(w, "internal server error", 500)
		return
	}

	if msg.SatelliteID == -1 || msg.APIKey == "" {
		log.Println(&Error{Name: remoteAddr, Message: "missing required message parameter"})
		http.Error(w, "internal server error", 500)
		return
	}

	if msg.TARBase64 == "" {
		log.Println(&Error{Name: remoteAddr, Message: "received export without data"})
		http.Error(w, "internal server error", 500)
		return
	}

	sio, ok := s.satelliteIO[msg.SatelliteID]
	if !ok {
		log.Println(&Error{Name: remoteAddr, Message: fmt.Sprint("unknown satellite id ", msg.SatelliteID)})
		http.Error(w, "internal server error", 500)
		return
	}

	if sio.APIKey() != msg.APIKey {
		log.Println(&Error{Name: remoteAddr, Message: "satellite auth failed"})
		http.Error(w, "internal server error", 500)
		return
	}

	err := nstautil.TarExtractDecode("/opt/openitc/nagios/etc", msg.TARBase64)
	if err != nil {
		log.Println(&Error{Name: remoteAddr, Err: err})
		http.Error(w, "internal server error", 500)
		return
	}

	// Runs the export on the satellite system (client_pull)
	_ = sio.Export(msg)

	lastExport := sio.LastExport()
	if err := json.NewEncoder(w).Encode(&httpproto.ExportData{
		SatelliteID: sio.SatelliteID(),
		APIKey:      sio.APIKey(),
		LastExport:  lastExport,
		TARBase64:   "",
	}); err != nil {
		log.Println(&Error{Name: "Local", Message: "could not marshal response", Err: err})
		http.Error(w, "internal server error", 500)
		return
	}
}

// Stop the webserver
func (s *NSTAHTTPServer) Stop() {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	if !s.stopped {
		log.Println(&Error{Name: "Local", Message: "begin with shutdown of webserver"})
		s.stopped = true
		to := time.NewTimer(time.Second * 5)
		defer to.Stop()

		ctx, cancel := context.WithCancel(context.Background())
		completeHTTP := make(chan struct{})
		completeHTTPS := make(chan struct{})
		completed := 0
		go func() {
			_ = s.serverHTTP.Shutdown(ctx)
			completeHTTP <- struct{}{}
		}()
		go func() {
			_ = s.serverHTTPS.Shutdown(ctx)
			completeHTTPS <- struct{}{}
		}()

		for {
			select {
			case <-completeHTTP:
				completed++
			case <-completeHTTPS:
				completed++
			case <-to.C:
				log.Println(&Error{Name: "Local", Message: "webserver shutdown timeout reached"})
				completed = 2
			}
			if completed >= 2 {
				break
			}
		}

		cancel()
	}
	log.Println(&Error{Name: "Local", Message: "wait for webserver shutdown to finish"})
	s.wg.Wait()
	log.Println(&Error{Name: "Local", Message: "webserver shutdown complete"})
}

// RegisterSatelliteIO adds a new server to the webservers callback list
func (s *NSTAHTTPServer) RegisterSatelliteIO(io SatelliteIO) {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	s.satelliteIO[io.SatelliteID()] = io
}

// UnregisterSatelliteIO removes a server from the webservers callback list
func (s *NSTAHTTPServer) UnregisterSatelliteIO(satelliteID int) {
	s.mtx.Lock()
	defer s.mtx.Unlock()

	delete(s.satelliteIO, satelliteID)
}

// New webserver for NSTA
func New(logWriter io.Writer, nagiosPath string, webserverAddressHTTP, webserverAddressHTTPS, certFile, keyFile string, wType WebserverType) *NSTAHTTPServer {
	s := &NSTAHTTPServer{
		nagiosPath:    nagiosPath,
		stopped:       false,
		webserverType: wType,
		satelliteIO:   make(map[int]SatelliteIO),
		certFile:      certFile,
		keyFile:       keyFile,
	}

	router := mux.NewRouter().StrictSlash(true)
	router.HandleFunc("/nsta/api/satcom", s.satcom).Methods("POST")

	switch s.webserverType {
	case WebserverTypeMaster:
		router.HandleFunc("/nsta/api/export", s.exportMaster).Methods("POST")
	case WebserverTypeSatellite:
		router.HandleFunc("/nsta/api/export", s.exportSatellite).Methods("POST")
	}

	handler := handlers.ContentTypeHandler(handlers.RecoveryHandler()(handlers.CombinedLoggingHandler(logWriter, router)), "application/json")

	s.serverHTTP = &http.Server{
		Addr:         webserverAddressHTTP,
		Handler:      handler,
		ReadTimeout:  time.Minute * 5,
		WriteTimeout: time.Minute * 5,
	}

	s.serverHTTPS = &http.Server{
		Addr:         webserverAddressHTTPS,
		Handler:      handler,
		ReadTimeout:  time.Minute * 5,
		WriteTimeout: time.Minute * 5,
		TLSConfig: &tls.Config{
			MinVersion: tls.VersionTLS12,
		},
	}

	return s
}

// Run should be called as go routine and starts the webserver
func (s *NSTAHTTPServer) Run() {
	s.wg.Add(2)

	go func() {
		if err := s.serverHTTP.ListenAndServe(); err != nil && err != http.ErrServerClosed {
			log.Fatalln(&Error{Name: "Local", Message: "could not start local webserver", Err: err})
		}
		log.Println(&Error{Name: "Local", Message: "all http connections stopped"})
		s.wg.Done()
	}()

	go func() {
		if err := s.serverHTTPS.ListenAndServeTLS(s.certFile, s.keyFile); err != nil && err != http.ErrServerClosed {
			log.Fatalln(&Error{Name: "Local", Message: "could not start local webserver", Err: err})
		}
		log.Println(&Error{Name: "Local", Message: "all https connections stopped"})
		s.wg.Done()
	}()
}
